import glob
import os
import dill
import datetime
from functools import wraps
from pytils.configurator import *
from pytils.logger import logger

# How long the object will be fresh? None - option for no usage of pickle. Can be change in config.
period = config_var_with_default('PICKLE_PERIOD_DEFAULT', 1)


def pickledays(period=period):
    def picklecache(func):
        # Path of storage the pickle files.
        path_pickle = config_var_with_default('PATH_PICKLE', './Assets/pickle/') + func.__name__

        def makename(*args, **kwargs):
            """naming the pickle file"""

            # define the function for stringify the arguments
            def convert_type(x):
                if isinstance(x, datetime.datetime):
                    return str(x.date())
                else:
                    # not so long names should be used
                    return str(x)[:20]

            name = ''
            # list through arguments and add them to file name
            for sublist in args:
                if isinstance(sublist, dict):
                    for key in sorted(sublist.keys()):
                        name += convert_type(key) + convert_type(sublist[key])
                else:
                    for key in sublist:
                        name += convert_type(key)
            if name == '':
                name = 'NA'
            return path_pickle + '/' + name

        def clearcache(*args, **kwargs):
            """ delete the cached result for these particular arguments """
            cachename = makename(args, kwargs)
            try:
                os.remove(cachename)

            except FileNotFoundError:
                pass

        def clearallcache():
            """ delete all chached results for this function """

            for f in glob.iglob(''.join(('.*', func.__name__, '_picklecache'))):
                try:
                    os.remove(f)
                except FileNotFoundError:
                    pass

        @wraps(func)
        def wrapper(*args, **kwargs):
            """wrapper which does the actual caching"""

            cachename = makename(args, kwargs)

            def write():
                # TODO pickle with import issues https://stackoverflow.com/questions/1412787/picklingerror-cant-pickle-class-decimal-decimal-its-not-the-same-object
                if not os.path.exists(path_pickle):
                    os.makedirs(path_pickle)

                result = func(*args, **kwargs)
                dill.dump(result, open(cachename, 'wb'))
                return result

            def read():
                with open(cachename, "rb") as f:
                    result = dill.load(f)
                return result

            try:
                import datetime
                ftime = (datetime.datetime.now() - datetime.datetime.fromtimestamp(os.path.getmtime(cachename)))
                if ftime.days > period or period is None:
                    logger.info('{} smell during {} > {}. Try to reload.'.format(func.__name__, ftime, period))
                    raise FileExistsError
                else:
                    logger.debug('{} fresh {}'.format(func.__name__, ftime))
                    result = read()
            except:
                # if file not founded and read unsuccessful
                result = write()
                logger.debug('{} refreshed'.format(func.__name__))
            return result

        # attach clearcache and clearallcache to wrapper
        wrapper.clearcache = clearcache
        wrapper.clearallcache = clearallcache

        return wrapper

    return picklecache



def Singleton_args(decorated_class):
    """Decorator for a class to make a singleton out of it.
    Have to be used with @Singleton_args before any class.

    Singleton - a software design pattern that restricts the instantiation of a class to one "single" instance.
    This is useful when exactly one object is needed to coordinate actions across the system.
    """
    class_instances = {}

    def getInstance(*args, **kwargs):
        """ creating or just return the one and only class instance.
            The singleton depends on the parameters used in __init__ """
        key = (decorated_class, args, str(kwargs))
        if key not in class_instances:
            class_instances[key] = decorated_class(*args, **kwargs)
        return class_instances[key]

    return getInstance



from pytils.logger import *
"""All tests here are not automated."""

@log('WARNING', True)
def func_tst(a=[1,3], b=2, c=3):
    return True

@log('WARNING', True)
def func_with_error(a=[1,3], b=2, c=3):
    raise ValueError('testError')


def test_log():
    answer = func_tst()
    # TODO check the print in StreamHandler and File
    assert answer


def test_log_with_args():
    answer = func_tst([11, 'beta'], 2, c=3)
    # TODO check the print in StreamHandler and File
    assert answer


def test_func_exception():
    try:
        answer = func_with_error([11, 'beta'], 2, c=3)
    except ValueError as ex:
        logger.exception('test accepted', ex)
        assert True

def test__la_exception():
    try:
        ValueError('sdf')
    except ValueError as ex:
        logger.error('test accepted', ex)

def test_colors():
    logger.debug("this is a debugging message")
    logger.success("this is a success message")
    logger.info("this is an informational message")
    logger.warning("this is a warning message")
    logger.notice("this is a notice message")
    logger.error("this is an error message")
    logger.critical("this is a critical message")
    logger.log(89, "this is a number message")


def test_loggers_conflict():
    logger.critical("this is a critical message")
    create_logger(__name__)
    logger2 = logging.getLogger(__name__)
    logger2.critical("this is a another critical message")


import pandas
from pandas import ExcelWriter

def create_excel_table(df, wr, sheet_name = 'Sheet1'):
    if not df.index.name:
        df.index.name = 'Index'

    # index and skip one row to allow us to insert a user defined header.
    if df.index.name not in df.columns:
        df = df.reset_index()
    df.to_excel(wr, sheet_name=sheet_name, startrow=1, header=False, index=False)

    # Get the xlsxwriter workbook and worksheet objects.
    worksheet = wr.sheets[sheet_name]

    # Get the dimensions of the dataframe.
    (max_row, max_col) = df.shape

    # Create a list of column headers, to use in add_table().
    column_settings = [{'header': column} for column in df.columns]

    # Add the Excel table structure. Pandas will add the data.
    worksheet.add_table(0, 0, max_row, max_col - 1, {'columns': column_settings, 'name': sheet_name})

    # Make the columns wider for clarity.
    worksheet.set_column(0, max_col - 1, 12)
    return wr



import json
import logging
from logging.handlers import TimedRotatingFileHandler
from functools import wraps
from socket import gethostname

import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning

from pytils.configurator import config_var_with_default

# totally reject the SSL check. Important information have to be logged without this module.
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
logging.getLogger("urllib3").setLevel(logging.WARNING)


def addLoggingLevel(levelName: str, levelNum: int, methodName=None):
    """
    Comprehensively adds a new logging level to the `logging` module and the
    currently configured logging class.

    `levelName` becomes an attribute of the `logging` module with the value
    `levelNum`. `methodName` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `methodName` is not specified, `levelName.lower()` is
    used.

    To avoid accidental clobberings of existing attributes, this method will
    raise an `AttributeError` if the level name is already an attribute of the
    `logging` module or if the method name is already present

    Example
    -------
     addLoggingLevel('TRACE', logging.DEBUG - 5)
     logging.getLogger(__name__).setLevel("TRACE")
     logging.getLogger(__name__).trace('that worked')
     logging.trace('so did this')
     logging.TRACE
    """
    if not methodName:
        methodName = levelName.lower()

    if hasattr(logging, levelName):
        raise AttributeError('{} already defined in logging module'.format(levelName))
    if hasattr(logging, methodName):
        raise AttributeError('{} already defined in logging module'.format(methodName))
    if hasattr(logging.getLoggerClass(), methodName):
        raise AttributeError('{} already defined in logger class'.format(methodName))

    # This method was inspired by the answers to Stack Overflow post
    # http://stackoverflow.com/q/2183233/2988730, especially
    # http://stackoverflow.com/a/13638084/2988730
    def logForLevel(self, message, *args, **kwargs):
        if self.isEnabledFor(levelNum):
            self._log(levelNum, message, args, **kwargs)

    def logToRoot(message, *args, **kwargs):
        logging.log(levelNum, message, *args, **kwargs)

    logging.addLevelName(levelNum, levelName)
    setattr(logging, levelName, levelNum)
    setattr(logging.getLoggerClass(), methodName, logForLevel)
    setattr(logging, methodName, logToRoot)

def create_logger(name = __name__):
    # agent = f"{__name__}Bot"
    logger = logging.getLogger(name)

    # define list of log handlers. Unified for further usage.
    # unpublic discord server have to be changed in config files.
    discord_channel = config_var_with_default("LOG_WEBHOOK_DISCORD",
                                              'https://discordapp.com/api/webhooks/748465782551216160/66Yn1W-PlVW5_PItHGxHMQ7ZRtkD37poEtIb9JeMlv3ricIgMEuyz17Sp1LtevDc0drl')
    logfile_path = config_var_with_default("LOG_FOLDER", './Assets/logs/')

    # Define level of allers
    discord_level = config_var_with_default("LOG_LEVEL_DISCORD", 'ERROR')
    logfile_level = config_var_with_default("LOG_LEVEL_FILE", 'ERROR')
    stream_level = config_var_with_default("LOG_LEVEL_STREAM", 'DEBUG')

    # Create DiscordHandlerand StreamHandler
    discord_handler = DiscordHandler(discord_channel)

    stream_handler = logging.StreamHandler()

    # Create FileHandler
    import os
    if not os.path.exists(logfile_path):
        os.makedirs(logfile_path)

    logfile_handler = TimedRotatingFileHandler(logfile_path + 'log', when='D', backupCount=14)

    # Set log level to handlers
    discord_handler.setLevel(discord_level)
    logfile_handler.setLevel(logfile_level)
    stream_handler.setLevel(stream_level)

    # Create formatter
    logs_format = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")

    # Add format to handlers
    discord_handler.setFormatter(DiscordFormatter())
    # discord_handler.setFormatter(logs_format)
    logfile_handler.setFormatter(logs_format)
    # stream_handler.setFormatter(logs_stream_format)

    # Add the handlers to the Logger
    logger.addHandler(discord_handler)
    logger.addHandler(logfile_handler)
    logger.addHandler(stream_handler)


    # add colors to stram logs
    import coloredlogs
    coloredlogs.install(level='DEBUG',
                        logger=logger,
                        level_styles={'debug': {'color': 95},
                                      'success': {'color': 46},
                                      'info': {'color': 'blue'},
                                      'notice': {'color': 'magenta'},
                                      'warning': {'color': 'yellow'},
                                      'error': {'color': 'red'},
                                      'critical': {'bold': True, 'color': 'red'}})

    logger.debug(f'Logger {name} set up')
    return logger


class DiscordFormatter(logging.Formatter):
    colormap = {'CRITICAL': 0xa11f1f, 'ERROR': 0xd10909,
                'WARNING': 0xb76b0d, 'NOTICE': 0x7c4605,
                'SUCCESS': 0x28a904, 'INFO': 0x0867af,
                'DEBUG': 0x676a6c
                }

    def formatMessage(self, record) -> dict:
        """
        Overwritten to return a dictionary of the relevant LogRecord attributes instead of a string.
        KeyError is raised if an unknown attribute is provided in the fmt_dict.
        Build the discord component https://autocode.com/tools/discord/embed-builder/
        """
        return {"embeds": [
                            {
                              "type": "rich",
                              "description": record.msg,
                              "color": self.colormap[record.levelname] if record.levelname in self.colormap else 0x181c20,
                              "timestamp": self.formatTime(record),
                              "footer": {
                                "text": f"{record.levelname} in {record.module}"
                              }
                            }]}

    def format(self, record) -> str:
        """
        Mostly the same as the parent's class method, the difference being that a dict is manipulated and dumped as JSON
        instead of a string.
        """
        message_dict = self.formatMessage(record)

        # TODO Format exception. Code below, but posts are too big
        # if record.exc_info:
        #     # Cache the traceback text to avoid converting it multiple times
        #     # (it's constant anyway)
        #     if not record.exc_text:
        #         message_dict["embeds"][0].update({"title": record.message,
        #                                           'description': self.formatException(record.exc_info)})
        return json.dumps(message_dict, default=str)



class DiscordHandler(logging.Handler):
    """
    A handler class which writes logging records, appropriately formatted, to a Discord Server using webhooks.
    Thx https://github.com/TrayserCassa/DiscordHandler
    """

    def __init__(self, webhook_url: str, agent=None):
        logging.Handler.__init__(self)

        if webhook_url is None or webhook_url == "":
            raise ValueError("webhook_url parameter must be given and can not be empty!")

        if agent is None or agent == "":
            agent = gethostname()

        self._url = webhook_url
        self._agent = agent
        self._header = self.create_header()
        self._name = ""

    def create_header(self):
        return {
            'User-Agent': self._agent,
            "Content-Type": "application/json"
        }

    def write_to_discord(self, message: str):
        try:
            #TODO use async thread for not waiting the answer from Discord
            request = requests.post(self._url,
                                    headers=self._header,
                                    data=message,
                                    verify=False,
                                    timeout=1)
        except requests.exceptions.ReadTimeout as ex:
            logger.debug('Discord logs timed out')
            raise ConnectionError('Discord timeout')
            # raise requests.exceptions.ReadTimeout

        if request.status_code == 404:
            raise requests.exceptions.InvalidURL(
                "This URL seams wrong... Response = %s" % request.text)

        if request.ok is False:
            raise requests.exceptions.HTTPError(
                "Request not successful... Code = %s, Message = %s" % request.status_code, request.text)

    def emit(self, record):
        try:
            msg = self.format(record)
            self.write_to_discord(msg)
        except Exception:
            self.handleError(record)


def log(level=None, arg_included=True):
    """Decorator for functions, which will log the function request.
    Have to be used with @log(level='YOUR LEVEL') before any function.
    """

    def log_without_level(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                if arg_included:
                    msg = "{0}: {1}, {2}".format(func.__name__, str(args), str(kwargs))
                else:
                    msg = func.__name__
                logger.debug(f"Processing {msg}", extra={'argi': args, 'kwargi': kwargs})
                res = func(*args, **kwargs)
                if level is None:
                    logger.success(msg)
                elif level == 'DEBUG':
                    logger.debug(msg)
                elif level == 'INFO':
                    logger.info(msg)
                elif level == 'WARNING':
                    logger.warning(msg)
                elif level == 'ERROR':
                    logger.error(msg)
                elif isinstance(level, int):
                    logger.log(msg, level=level)
                else:
                    raise AttributeError('Error for @log decorator arguments')
            except Exception as ex:
                logger.exception("{0}: {1}, {2} \n {3}".format(func.__name__, str(args), str(kwargs), ex))
                raise ex
            return res

        return wrapper

    return log_without_level


# add log levels into logging module
addLoggingLevel('SUCCESS', 15, methodName=None)
addLoggingLevel('NOTICE', 25, methodName=None)


#create one logger for reserve goals. Just import module with "from pytils.logger import logger" and use in your programm
create_logger("|")
logger = logging.getLogger("|")



import pytest

from pytils.singleton import Singleton_args


@Singleton_args
class Example:
    """Text class with few operations, which will be used in test cases."""

    def __init__(self, par=0):
        # Default value will be used for cases where args are not necessery
        self.par = par

    def __add__(self, other):
        self.par += other
        return self.par

    def __eq__(self, other):
        if isinstance(other, int):
            return self.par == other
        elif hasattr(self, 'par'):
            return self.par == other.par
        else:
            return False

    def __str__(self):
        return str(self.par)


num_list = [1, 2, 5]


@pytest.mark.parametrize("num", num_list)
def test_singlon_class_with_args(num):
    test_object1 = Example(num)
    assert test_object1 == num

    test_object2 = Example(num)
    assert test_object2 == num
    assert test_object2 == test_object1

    test_object1 += 2
    assert test_object1 == num + 2
    assert test_object2 == num + 2
    assert test_object2 == test_object1


def test_singlon_class_without_args():
    test_object1 = Example()
    assert test_object1 == 0

    test_object2 = Example()
    assert test_object2 == 0
    assert test_object2 == test_object1

    test_object1 += 2
    assert test_object1 == 2
    assert test_object2 == 2
    assert test_object2 == test_object1



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pytils-functions",
    version="0.1.1",
    author="Whispered",
    author_email="bluden99@example.com",
    description="Utils for data python project",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Whisperes/pytils",
    packages=setuptools.find_packages(),
    install_requires=['requests','dill','coloredlogs', 'pytest'],
    # TODO: from pip.req import parse_requirements
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


def config_var_with_default(var: str, default):
    """Create variable in settings (config) file if it not exist with defined measure.
    If variable already exists on project level, then take the parametr value from file. Value can be changed by user.
    """
    try:
        from config import settings
        if var not in settings:
            from dynaconf.loaders.toml_loader import write
            write('settings.toml', settings_data={var: default}, merge=True)
            return default
        else:
            return settings[var]
    except:
        return default
